Add LoadPath "../Support".
Add LoadPath "../Common".
Require Import List Arith.
Require Import Prelude Permission GenericTrace.
Require Import Trace.
Require Import Trace_Basics_projection.
Require Import Trace_Basics_occurences.
Require Import Trace_Basics_range.
Require Import Trace_Basics_father.
Require Import Trace_Basics_owns.
Require Import Trace_Basics_tribe.


Module Make (Perm : Permission)
            ( Export Address: DecidableInfiniteSet) 
            ( Export T : Type_.TYPE Address )
            ( Export V : Value.TYPE Address T ) 
            ( Tr : TraceT Perm Address T V)
            ( P : Proj Perm Address T V Tr)
            ( O : OccurencesT Perm Address T V Tr P)
            ( F : FatherT Perm Address T V Tr P O) 
            ( OW : OwnsT Perm Address T V Tr P O)
            ( R : RangeT Perm Address T V Tr P O)
            ( Tribe : TribeT Perm Address T V Tr P O F OW R).

  Import Tr P O F OW R Tribe.

  (** pending *)
  
  Lemma pending_se_s : 
    forall s t a,
      wf_occurences (s • (t,a)) ->
      wf_open_close (s • (t,a)) ->
      forall p,
      pending (s • (t,a)) p -> a <> Open p -> pending s p.
  Proof.
    intros s t a h_wf_occurences h_wf_open_close p h_pending h_neq.
    unfold pending.
    destruct h_pending as [[i h_i] h_not_occursIn].
    autorewrite with length in h_i; simpl in h_i.
    replace (length s + 1 - 1) with (length s) in h_i by omega. 
    split.
    - (exists i).
      destruct (range_se_s_neq_open _ _ _ h_wf_occurences h_wf_open_close p i (length s) h_i h_neq).
      + elim h_not_occursIn; exists (length s); destruct H; eauto with nth_error.
      + tauto.
    - contradict h_not_occursIn.
      eauto with occurences.
  Qed.

  (** - a section [p] pending in [s] is also pending in [s • (t,a)] if [a <> Close p] *)
  Lemma pending_s_se : 
    forall s t a,
      wf_occurences (s • (t,a)) ->
      wf_open_close (s • (t,a)) ->
      forall p, pending s p -> a <> Close p -> pending (s • (t,a)) p.
  Proof.
    intros s t a h_wf_occurences h_wf_opend_close p h_pending h_neq.
    destruct h_pending as [[i h_i] h_not_occursIn].
    split.
    - (exists i).
      destruct (range_s_se s p i (length s - 1) (t,a) h_i).
      destruct H.
      contradict h_not_occursIn.
      destruct s; simpl in *.
      + inversion h_i; subst; autorewrite with nth_error in Hi; discriminate.
      + replace (length s - 0) with (length s) in * by omega.
      exists (length s).
      assert (length s  < length (p0::s)) by intuition.
      eauto with nth_error.
      + autorewrite with length; simpl.
        now replace (length s + 1 - 1) with (length s) by omega.
    - contradict h_not_occursIn; eauto with occurences.
  Qed.

  
  (** - a section freshly opened is pending *)
  Lemma pending_open : 
    forall  s t p,
      wf_occurences (s • (t,Open p)) ->
      wf_open_close (s • (t,Open p)) ->
      pending (s • (t, Open p)) p.
  Proof.
    intros s t p h_wf_occurences h_wf_open_close.
    assert (~ occursIn (s • (t,Open p)) (Close p)).
    { 
      intro h_occursIn.
      inversion h_occursIn as [i ? h_i]; subst.
      assert (exists j, j < i /\ action_of (pi j (s • (t, Open p))) == Open p) as [j [h_a h_b]].
      {
        destruct h_wf_open_close with i p as [j [h_a [h_b _]]]; try assumption.
        exists j; tauto.
      }
      assert (action_of (pi (length s) (s • (t,Open p))) == Open p)
        by (autorewrite with nth_error; reflexivity).
      replace j with (length s) in * by wellFormed_occurences (Open p).
        assert (i < length (s • (t,Open p))) by eauto with nth_error.
        autorewrite with length in H0; simpl in H0; intuition.
    }
    split.
    - (exists (length s)).
      constructor 2; [autorewrite with nth_error; reflexivity | assumption].
    - assumption.
  Qed.

  Lemma range_last_se_pending_s :
    forall s e p i,
      wellFormed (s • e) ->
      range (s • e) p i (length s) -> i <> length s -> pending s p.
  Proof.
    intros s e p i h_wf h_range h_neq.
    assert (i < length (s • e)) as Hi_se.
    {
      inversion h_range; subst;
      eapply lift_nth_error_defined_left;
      exists(Open p); eassumption.
    }
    assert (i < length s) as Hi_s 
           by (autorewrite with length in * ; simpl in *; intuition).
    assert (action_of (pi i (s • e)) == Open p) as Ha_i_se
           by now inversion h_range.
    assert (action_of (pi i s) == Open p) as Ha_i_s
           by (erewrite <- lift_nth_error_append_left; eassumption).
    assert (~ occursIn s (Close p)).
    {
      intro H.
      inversion h_range as [ ? Haction ?| ? H1 H2 H3 H4]; subst.
      - inversion H as [i0 ?]; subst.
        assert(i0 < length s) as Hi0 by eauto with nth_error.
        assert(i0 = length s) as Hi0_eq.
        {
          assert (action_of (pi i0 (s • e)) == Close p)  by eauto with nth_error.
          wellFormed_occurences (Close p).
        }
        rewrite Hi0_eq in Hi0.
        elim (Lt.lt_irrefl (length s) Hi0).
      - elim H2.
        now apply occursIn_s_se.
    }
    split.
    exists i.
    constructor 2; assumption.
    assumption.
  Qed.

  Lemma pending_range_last :
    forall s p e,
      pending s p -> exists i, range (s • e) p i (length s).
  Proof.
    intros s p e H.
    inversion H as [[i Hrange] H2].
    inversion Hrange as [ ? Haction ? | ? Haction Hoccurs]; subst;
    assert(i < length s) as Hi_len by eauto with nth_error.
    - exists i.
      constructor.
      + now rewrite lift_nth_error_append_left.
      + contradict H2.
        eauto.
    - exists i.
      assert(exists t', threadId_of(pi i s) == t') as Ht'.
      {
        assert(exists y,pi i s==y) as Hy 
          by (now apply ListBasics.nth_error_lt_defined).
        destruct Hy as [ [t' a'] Ht'a'].
        exists t'.
        rewrite Ht'a'.
        trivial.
      }
      destruct Ht' as [t' Ht'].
      destruct(occursIn_dec (s • e) (Close p)) as [Ho | Ho].
      + inversion Ho as [i0 ? Haction_i0]; subst.
        assert(action_of (pi i (s • e)) == Open p) 
          by (now rewrite lift_nth_error_append_left).
        assert(i0 = length s).
        {
          destruct(lt_eq_lt_dec i0 (length s)) as [[Hi0|Hi0]|Hi0].
          - contradict Hoccurs.
            assert(action_of(pi i0 s)==Close p) by 
                (now rewrite lift_nth_error_append_left in Haction_i0).
            econstructor; eassumption.
          - trivial.
          - assert(i0>=length (s • e)) by 
                (autorewrite with length; simpl; intuition).
            assert(action_of (pi i0 (s • e)) = None) as Hnone by
                (now rewrite ListBasics.nth_errorGeLength).
            rewrite Hnone in Haction_i0.
            discriminate.
        }
        subst.
        now constructor.
      + assert(action_of (pi i (s • e)) == Open p) 
          by (now rewrite lift_nth_error_append_left).
        assert(length s = length(s • e) - 1) as Hs
          by (autorewrite with length; simpl; intuition).
        rewrite Hs.
        now constructor 2.
  Qed.
  
  Lemma pending_range_last2 : 
    forall s p, 
      pending s p ->
      exists i, range s p i (length s - 1).
  Proof.
    intros s p H.
    inversion H as [[i Hrange] H2].
    exists i.
    trivial.
  Qed.

  (** exclude *)
  
  Lemma exclude_nil : forall t p, ~ exclude nil p t.
  Proof.
    intros t p h_exclude.  
    destruct h_exclude.
    inversion H.
    inversion H1.
    inversion H3; subst.
    destruct x; discriminate.
    destruct x; discriminate.
  Qed. 
  
  Lemma exclude_se_s_not_open_p : 
    forall s p t (HWF : wellFormed (s • (t,Open p))) p' t',
      exclude (s • (t,Open p)) p' t' -> p <> p' -> exclude s p' t'. 
  Proof.
    intros.
    inversion H.
    constructor.
    inversion HWF.
    apply pending_se_s with (t:=t0) (a:=Open p).
    assumption.
    assumption.
    assumption.
    intro.
    injection H3; intuition.
    intro; elim H2.
    apply tribe_s_se.
    assumption.
  Qed.

  Lemma exclude_se___s_not_open : 
    forall s p t t' a (HWF : wellFormed (s •(t,a))), 
      exclude (s • (t,a)) p t' -> a <> Open p -> exclude s p t'. 
  Proof.
    intros s p t0 t' a HWF H HnotOpen.
    destruct H as [Hpending Htribe].
    split.
    - (* pending *)
      destruct Hpending as [[i Hrange] HnotOccursIn].
      assert(not(occursIn s (Close p))) by (contradict HnotOccursIn; 
                                            now apply occursIn_s_se).
      inversion Hrange as [ ? Haction ? | ? Haction Hoccurs]; subst.
      + contradict HnotOccursIn.
        eauto.
      + assert(i < length s).
        {
          destruct(lt_eq_lt_dec i (length s)) as [[Hi|Hi]|Hi].
          - trivial.
          - subst. 
            rewrite ListBasics.nth_error_append_cons_eq in Haction.
            inversion Haction.
            contradict HnotOpen.
            trivial.
          - assert(i>=length (s • (t0, a))) 
              by (autorewrite with length; simpl;intuition).
            rewrite ListBasics.nth_errorGeLength in Haction; trivial.
            discriminate.
        }
        assert(action_of (pi i s) == Open p) by
            (now rewrite lift_nth_error_append_left in Haction).
        assert(range s p i (length s - 1)) by (now constructor 2).
        constructor; eauto.
    - (* tribe *) 
      contradict Htribe.
      now apply tribe_s_se.
  Qed.

(*Lemma exclude_s_se_not_close : 
    forall s p t,
      exclude s p t -> forall t' a', 
                        a' <> Close p -> exclude (s • (t',a')) p t.
  Proof.
    intros s p t H t' a' HnotClose.
    destruct H as [Hpending Htribe].
    split.
    - (* pending *)
      destruct Hpending as [[i Hrange] HnotOccursIn].
      inversion Hrange as [ ? Haction ? | ? Haction Hoccurs]; subst.
      + contradict HnotOccursIn.
        eauto.
      + assert(not(occursIn (s • (t',a')) (Close p))) 
          by  (contradict HnotOccursIn; eapply occursIn_se_s_neq; eauto).
        assert(action_of (pi i ( s • (t', a'))) == Open p) 
          by (assert(i < length s) by (eauto with nth_error);
              eauto with nth_error).
        assert(range (s • (t', a')) p i (length (s • (t', a')) - 1))
          by (now apply range_opened).
        constructor; eauto.
    - (* tribe *)
  False: additional condition necessary

  Lemma exclude_s_se : 
    forall s p t a, (wellFormed (s • (t,a))) -> 
                    ~ exclude (s • (t,a)) p t -> ~ exclude s p t.
  Proof.
    intros s p t a H H0 HExclude.
    elim H0.
    apply exclude_s_se_not_close.
    assumption.
    intro; subst.
    assert (action_of (pi (length s) (s •(t,Close p))) == Close p) 
           by (autorewrite with nth_error;trivial).
    inversion H.
    destruct (WF4 _ _ H1) as [i [HLt [Hc Hd]]].
    assert (action_of (pi i s) == Open p) 
           by (erewrite <- lift_nth_error_append_left; eassumption).
    assert (threadId_of (pi i s) == t).
    {
      specialize (WF4 (length s) p H1).
      destruct WF4 as [j [Hj_lt [ Hj_a Hj_t] ] ].
      assert(j = i) by (eapply WF1 with (a:=Open p); try eassumption; constructor).
      subst.
      assert(length s<length (s •(t,Close p)))by(autorewrite with length;simpl;intuition).
      rewrite lift_nth_error_append_left with (k:=i) in Hj_t; trivial.
      rewrite <- Hj_t.
      autorewrite with nth_error.
      trivial.
    }
    assert (tribe s p t).
    constructor 1.    
    apply owns_cons with (i:=i).
    assumption.
    assumption.
    destruct HExclude.
    elim H6.
    assumption.
  Qed.
 *)

  Lemma exclude_s_open_not_owner :
    forall s p t,
      wellFormed (s • (t,Open p)) ->
      forall t',
        t' <> t -> exclude (s • (t, Open p)) p t'.
  Proof.
    intros s p t h_wf_formed t' h_neq.
    inversion h_wf_formed.
    split; eauto using pending_open, notInTribe_s_open_not_owner.
  Qed.

(*  Lemma outerExclude_close : 
    forall s t' p (HWF : wellFormed (s • (t', Close p))) t,
      outerExclude s p t  ->  forall p', ~ exclude (s • (t',Close p)) p' t.
  Proof.
    intros s t' p HWF t Houter p' Hexclude_se.
    destruct(Perm.eq_dec p p') as [Heq|Hneq]; subst.
    - inversion Hexclude_se as [ Hpend _ ] .
      inversion Hpend as [ _ HoccursIn ].
      contradict HoccursIn.
      assert(action_of (pi (length s) (s • (t', Close p'))) == Close p') by
            (rewrite ListBasics.nth_error_append_cons_eq; trivial).
      eauto.
    - inversion Houter as [ Hexclude_s Hleftmost ].
      inversion HWF as [ WF1 _ _ _ _ _ _ HopenInSection _].
      inversion Hexclude_se as [ Hpend_se HnotTribe_se ].
      inversion Hexclude_s as [ Hpend HnotTribe ].
      inversion Hpend as [ Hrange HoccursIn ].
      destruct Hrange as [i Hrange_i].
      inversion Hpend_se as [ Hrange_se HoccursIn_se ].
      destruct Hrange_se as [j Hrange_j].
      inversion Hrange_i as [ H_i H_i' | H_i H_i']; subst.
      + inversion Hrange_j as [ H_j H_j' | H_j H_j' ]; subst.
        * assert(range (s • (t', Close p)) p i (length s)) as Hrange_se_p.
          {
            assert(i < length s) by (eauto with nth_error).
            assert(action_of (pi i (s • (t', Close p))) == Open p) by eauto with nth_error.
            assert(action_of (pi (length s) (s • (t', Close p))) == Close p) by 
                  (autorewrite with nth_error; trivial).
            now constructor.
          }
          assert(i < j). eapply occursAfterSingle; eauto. 
          econstructor; eauto.

          (* Hleftmost. ; try eassumption. *)
          (* specialize (HopenInSection p i (length s) p' j Hrange_se_p H_j'). *)
          (* edestruct HopenInSection. *)
  A_dmitted.*)

  (* A voir *)

(*  Lemma outerExclude_not_close : 
    forall s (HWF : wellFormed s) p t,
      outerExclude s p t -> 
      forall t' a', a' <> Close p -> (*a'<> Fork t ->*) outerExclude (s • (t',a')) p t.
  A_dmitted.*)

  Lemma outerExclude_open : 
    forall s p t,
      outerExclude s p t -> exists i, action_of (pi i s) == Open p.
  Proof.
    intros s p t0 H.
    inversion H as [ Hexclude HoccursAfter ].
    destruct Hexclude as [ Hpend Htribe ].
    inversion Hpend as [ Hrange ].
    destruct Hrange as [i Hi].
    inversion Hi; subst; eauto.
  Qed.

  Fact outerExclude_not_in_tribe : 
    forall s p t,
      outerExclude s p t -> ~ tribe s p t.
  Proof.
    intros s p t0 h_outerExclude. 
    inversion h_outerExclude.
    unfold exclude in H0.
    destruct H0.
    assumption.
  Qed.

  Lemma tribeChildren_live_after_open : 
    forall s, 
      wellFormed s ->
      forall p t,
        tribeChildren s p t ->
        forall i k ,
          action_of (pi i s) == Open p ->
          threadId_of (pi k s) == t ->
          i < k.
  Proof.
    intros s h_wf p t h_tc.
    assert (wf_fork s) as h_wf_fork by now inversion h_wf.
    induction h_tc.
    - intros i0 k0 h_open_p h_k_0_tid.
      assert (i0 = i) by (inversion H; subst; wellFormed_occurences (Open p)).
      assert (k < k0) by (eapply h_wf_fork; eauto).
      intuition.
    - intros i0 k0 h_open_p h_k_0_tid.
      destruct H as [k [Ha Hb]].
      assert (i0 < k) by eauto.
      assert (k < k0) by (eapply h_wf_fork; eauto).
      intuition.
  Qed.

 


 
  

End Make.